package ap.basetypes;

import ap.util.{Debug, Logic}

object MultiSet {
  
  private val AC = Debug.AC_BASE_TYPE
  
  def empty[A] = new MultiSet[A](scala.collection.immutable.HashMap.empty)
  
  private def addMultiplicity[A](multiplicities : Map[A, Int], el : (A, Int))
                                                            : Map[A, Int] = {
    val (key, mult) = el
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(MultiSet.AC, mult >= 0)
    ////////////////////////////////////////////////////////////////////////////
    if (mult == 0)
      multiplicities
    else
      multiplicities + (key -> (mult + multiplicities.getOrElse(key, 0)))    
  }

  private def removeMultiplicity[A](multiplicities : Map[A, Int], el : (A, Int))
                                                            : Map[A, Int] = {
    val (key, mult) = el
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(MultiSet.AC, mult >= 0)
    ////////////////////////////////////////////////////////////////////////////
    if (mult == 0) {
      multiplicities
    } else {
      val newMult = multiplicities.getOrElse(key, 0) - mult 
      if (newMult > 0)
        multiplicities + (key -> newMult)
      else
        multiplicities - key
    }    
  }
}

/**
 * A simple class to represent multisets
 */
class MultiSet[A] private (private val multiplicities : Map[A, Int])
      extends ((A) => Int) {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(MultiSet.AC,
                   Logic.forall(for ((key, mult) <- multiplicities.elements)
                                yield mult > 0))
  //////////////////////////////////////////////////////////////////////////////
  
  def apply(key : A) : Int = multiplicities.getOrElse(key, 0)
  
  def contains(key : A) : Boolean = multiplicities contains key
  
  lazy val size : Int = (0 /: multiplicities) ((res, n) => res + (n _2))

  lazy val supp : scala.collection.Set[A] = multiplicities.keySet
  
  //////////////////////////////////////////////////////////////////////////////

  private def updateMultiplicities(newMults : Map[A, Int]) : MultiSet[A] =
    if (multiplicities eq newMults)
      this
    else
      new MultiSet(newMults)
  
  //////////////////////////////////////////////////////////////////////////////

  def +(el : (A, Int)) : MultiSet[A] =
    updateMultiplicities(MultiSet.addMultiplicity(multiplicities, el))

  def +(el : A) : MultiSet[A] = this.+(el, 1)

  def ++(els : Iterator[(A, Int)]) : MultiSet[A] =
    updateMultiplicities((multiplicities /: els) ((m, el) =>
                                MultiSet.addMultiplicity(m, el)))

  def ++(els : Iterable[(A, Int)]) : MultiSet[A] =
    ++(els.elements)

  def +++(els : Iterator[A]) : MultiSet[A] =
    ++(for (el <- els) yield (el, 1))

  def +++(els : Iterable[A]) : MultiSet[A] =
    ++(for (el <- els.elements) yield (el, 1))

  def ++(els : MultiSet[A]) : MultiSet[A] =
    ++(els.multiplicities.elements)

  //////////////////////////////////////////////////////////////////////////////
    
  def -(el : (A, Int)) : MultiSet[A] =
    updateMultiplicities(MultiSet.removeMultiplicity(multiplicities, el))

  def --(els : Iterator[(A, Int)]) : MultiSet[A] =
    updateMultiplicities((multiplicities /: els) ((m, el) =>
                                MultiSet.addMultiplicity(m, el)))

  def --(els : Iterable[(A, Int)]) : MultiSet[A] = --(els.elements)

  def ---(els : Iterator[A]) : MultiSet[A] =
    --(for (el <- els) yield (el, 1))

  def ---(els : Iterable[A]) : MultiSet[A] =
    --(for (el <- els.elements) yield (el, 1))

  def --(els : MultiSet[A]) : MultiSet[A] = --(els.multiplicities.elements)

  //////////////////////////////////////////////////////////////////////////////

  override def toString : String = multiplicities.toString
}
