package ap.proof;

import ap.terfor.TermOrder
import ap.util.Debug

object ProofTreeOneChild {
  
  private val AC = Debug.AC_PROOF_TREE
  
  def unapply(p : ProofTree) : Option[ProofTree] = p match {
    case p : ProofTreeOneChild => Some(p.subtree)
    case _ => None
  }
  
}

/**
 * Common superclass for proof trees that have exactly one direct subtree. Such
 * trees know about two <code>TermOrder</code>s: the <code>TermOrder</code> of
 * the closing constraint coming from the subtree, and the
 * <code>TermOrder</code> of the constraint of this <code>ProofTree</code>
 */
trait ProofTreeOneChild extends ProofTree {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(ProofTreeOneChild.AC,
                   subtreeOrder isSortingOf subtree.closingConstraint)
  //////////////////////////////////////////////////////////////////////////////
   
  val subtree : ProofTree

  lazy val subtrees : Seq[ProofTree] = List(subtree)

  lazy val stepPossible : Boolean = subtree.stepPossible

  /**
   * Replace the subtree and the constant freedom status with new values.
   */
  def update(newSubtree : ProofTree,
             newConstantFreedom : ConstantFreedom) : ProofTree
  
  /**
   * Given a new constant freedom for this proof tree, derive the corresponding
   * freedom for the direct subtree.
   */
  def newConstantFreedomForSubtree(cf : ConstantFreedom) : ConstantFreedom
  
  protected val subtreeOrder : TermOrder
  
}
