package ap.proof;

import ap.proof.goal.Goal
import ap.terfor.linearcombination.LinearCombination
import ap.terfor._
import ap.util.{Debug, Logic, PlainRange, FilterIt, Seqs}

object TestProofTree {

  def assertNormalisedTree(t : ProofTree) : Unit =
    assert(isNormalisedTree(t) &&
           correctBindings(t, BindingContext.EMPTY) &&
           t.fixedConstantFreedom)

  /**
   * The information about bound constants is consistent with the actual
   * nodes in the proof tree
   */
  private def correctBindings(t : ProofTree, bc : BindingContext) : Boolean =
    (t match {
       case t : Goal => t.facts.isFalse || t.bindingContext == bc
       case _ => t.bindingContext == bc
    }) &&
    (t match {
       case QuantifiedTree(q, consts, subtree) =>
         correctBindings(subtree, bc.addAndContract(consts, q))
       case t : ProofTreeOneChild =>
         correctBindings(t.subtree, bc)
       case t : AndTree =>
         Logic.forall(for (subtree <- t.subtrees.elements)
                      yield correctBindings(subtree, bc))
       case _ => true
     })
  
  /**
   * The proof is fully expanded
   */
  private def isNormalisedTree(t : ProofTree) : Boolean = t match {
    case t : WeakenTree => t.disjunct.isTrue || isNormalisedTree(t.subtree)
    case t : ProofTreeOneChild => isNormalisedTree(t.subtree)
    case t : AndTree => Logic.exists(for (subtree <- t.subtrees.elements)
                                     yield isNormalisedTree(subtree) &&
                                           subtree.closingConstraint.isFalse) ||
                        Logic.forall(for (subtree <- t.subtrees.elements)
                                     yield isNormalisedTree(subtree))
    case goal : Goal =>
      goal.tasks.isEmpty &&
      (goal.facts.isFalse ||
        Logic.forall(for (lc <- goal.facts.arithConj.positiveEqs.elements)
                     yield isNormalisedPosEq(lc, goal)) &&
        Logic.forall(for (lc <- goal.facts.arithConj.negativeEqs.elements)
                     yield isNormalisedNegEq(lc, goal)) &&
        goal.facts.arithConj.inEqs.equalityInfs.isTrue &&
        Seqs.disjoint(goal.facts.arithConj.inEqs.constants,
                      goal.eliminatedConstants))
  }
  
  private def isNormalisedPosEq(lc : LinearCombination, goal : Goal) =
    lc.leadingCoeff.isOne
/*    if (goal eliminates lc.leadingTerm)
      (lc.leadingCoeff.isOne ||
       Logic.forall(1, lc.size, (i) => !(goal eliminates (lc(i) _2))))
    else
      lc.leadingCoeff.isOne */

  private def isNormalisedNegEq(lc : LinearCombination, goal : Goal) =
    Logic.forall(for (posLc <- goal.facts.arithConj.positiveEqs.elements)
                 yield ((lc.leadingTerm != posLc.leadingTerm) &&
                        (if (goal eliminates posLc.leadingTerm)
                           lc.get(posLc.leadingTerm) isAbsMinMod posLc.leadingCoeff
                         else
                           doesNotContain(lc, posLc.leadingTerm))))

  private def doesNotContain(tf : TerFor, t : Term) : Boolean = t match {
    case t : ConstantTerm => !(tf.constants contains t)
    case t : VariableTerm => !(tf.variables contains t)
    case _ => true
  }

}
