package ap.proof.goal;

import ap.terfor.conjunctions.Conjunction

/**
 * Trait for tasks that are 
 * given a <code>priority</code>, based on the age (to ensure fairness)
 * and on other criteria to work at different tasks in a meaningful order.
 * Tasks with smaller <code>priority</code> are supposed to be handled first.
 * Proof tasks of this kind cannot fail, i.e., the result of <code>apply</code>
 * is simply <code>ProofTree</code>.
 */
trait PrioritisedTask extends Task {

  val priority : Int
 
  /**
   * Update the task with possibly new information from the goal. If new facts
   * can be derived, these are put into <code>factCollector</code>
   */
  def updateTask(goal : Goal, factCollector : Conjunction => unit)
                                                   : Seq[PrioritisedTask]
  
}
